//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CluelessPlus - an noai-AI                                       //
//                                                                  //
//////////////////////////////////////////////////////////////////////
// 
// Author: Zuu (Leif Linse), user Zuu @ tt-forums.net
// Purpose: To play around with the noai framework.
//               - Not to make the next big thing.
// Copyright: Leif Linse - 2008, 2009
// License: GNU GPL - version 2

class Helper {
	company_name = null; // a copy of state_company_name which does not include the state-prefix that is included if one ask company.GetCompanyName();

	constructor() {
		company_name = "";
	}

	static function IsStraight(tile1, tile2);
	static function Distance(a, b);
	static function TownDistance(town1, town2);
	static function StepFunction(t);
	static function ArrayFind(array, toFind);
	static function TileLocationString(tile);
	static function SetSign(tile, message);
	static function BreakPoint(sign_tile);
	static function HasSign(text);
	static function ClearAllSigns();
	static function IsTownInConnectionList(connection_list, check_town);
	static function Debug(text);

	static function MakeTileRectAroundTile(center_tile, radius);
	static function GetTileRelative(relative_to_tile, delta_x, delta_y);
	static function GetTileRelativeClampToMap(relative_to_tile, delta_x, delta_y);

	static function ItemValuator(item);
	static function ListValueSum(ai_list);
	static function CopyListSwapValuesAndItems(old_list);
	static function GetListMinValue(ai_list);
	static function GetListMaxValue(ai_list);

	static function Clamp(value, min, max);
	static function Min(a, b);
	static function Max(a, b);
	static function Abs(a);

	static function CanConnectToRoad(road_tile, adjacent_tile_to_connect);
	static function GetNeighbours4(tile_id);
}

function Helper::TownDistance(town1, town2)
{
	return AIMap.DistanceManhattan(AITown.GetLocation(town1), AITown.GetLocation(town2));
}

function Helper::Distance(a, b)
{
	return AIMap.DistanceManhattan(a, b);
}

function Helper::StepFunction(t)
{
	if(t>=0)
		return 1;

	return 0;
}

function Helper::TileLocationString(tile)
{
	return "(" + AIMap.GetTileX(tile) + ", " + AIMap.GetTileY(tile) + ")";
}

function Helper::SetSign(tile, message)
{
	if(AIController.GetSetting("debug_signs") != 1)
		return;

	local found = false;
	local sign_list = AISignList();
	for(local i = sign_list.Begin(); sign_list.HasNext(); i = sign_list.Next())
	{
		if(AISign.GetLocation(i) == tile)
		{
			if(found)
				AISign.RemoveSign(i);
			else
			{
				if(message == "")
					AISign.RemoveSign(i);
				else
					AISign.SetName(i, message);
				found = true;
			}
		}
	}

	if(!found)
		AISign.BuildSign(tile, message);
}

// Places a sign on tile sign_tile and waits until the sign gets removed
function Helper::BreakPoint(sign_tile)
{
	if(AIController.GetSetting("debug_signs") != 1)
		return;

	if(Helper.HasSign("no_break"))
		return;

	AILog.Warning("Break point reached. -> Remove the \"break\" sign to continue.");
	local sign = AISign.BuildSign(sign_tile, "break");
	while(AISign.IsValidSign(sign)) { AIController.Sleep(1); }
}

function Helper::HasSign(text)
{
	local sign_list = AISignList();
	for(local i = sign_list.Begin(); sign_list.HasNext(); i = sign_list.Next())
	{
		if(AISign.GetName(i) == text)
		{
			return true;
		}
	}
	return false;
}
function Helper::ClearAllSigns()
{
	local sign_list = AISignList();
	for(local i = sign_list.Begin(); sign_list.HasNext(); i = sign_list.Next())
	{
		AISign.RemoveSign(i);
	}
}

function Helper::IsStraight(tile1, tile2)
{
	if(AIMap.GetTileX(tile1) == AIMap.GetTileX(tile2) ||
			AIMap.GetTileY(tile1) == AIMap.GetTileY(tile2))
	{
		return true;
	}
	return false;
}

function Helper::IsTownInConnectionList(connection_list, check_town)
{
	foreach(val in connection_list)
	{
		foreach(town in val.town)
		{
			if(town == check_town)
			{
				return true;
			}
		}
	}
	return false;
}

function Helper::Debug(text)
{
	//print(company_name + ": " + text);
	AILog.Info(text);
}

// RETURN null if not found, else the key to the found value.
function Helper::ArrayFind(array, toFind)
{
	
	foreach(key, val in array)
	{
		if(val == toFind)
		{
			return key;
		}
	}
	return null;
}

function Helper::GetPAXCargo()
{
	local cargo_list = AICargoList();
	cargo_list.Valuate(AICargo.HasCargoClass, AICargo.CC_PASSENGERS);
	cargo_list.KeepValue(1);
	cargo_list.Valuate(AICargo.GetTownEffect);
	cargo_list.KeepValue(AICargo.TE_PASSENGERS);

	if(!AICargo.IsValidCargo(cargo_list.Begin()))
	{
		AILog.Error("PAX Cargo do not exist");
	}

	//AILog.Info("PAX Cargo is: " + AICargo.GetCargoLabel(cargo_list.Begin()));

	return cargo_list.Begin();
}

function Helper::MakeTileRectAroundTile(center_tile, radius)
{
	local tile_x = AIMap.GetTileX(center_tile);
	local tile_y = AIMap.GetTileY(center_tile);

	AILog.Info("TileRectAroundTile: tile " + tile_x + ", " + tile_y);

	local x_min = Helper.Clamp(tile_x - radius, 1, AIMap.GetMapSizeX() - 2);
	local x_max = Helper.Clamp(tile_x + radius, 1, AIMap.GetMapSizeX() - 2);
	local y_min = Helper.Clamp(tile_y - radius, 1, AIMap.GetMapSizeY() - 2);
	local y_max = Helper.Clamp(tile_y + radius, 1, AIMap.GetMapSizeY() - 2);

	AILog.Info("TileRectAroundTile: x_min: " + x_min);
	AILog.Info("TileRectAroundTile: x_max: " + x_max);
	AILog.Info("TileRectAroundTile: y_min: " + y_min);
	AILog.Info("TileRectAroundTile: y_max: " + y_max);

	local list = AITileList();
	list.AddRectangle( AIMap.GetTileIndex(x_min, y_min), AIMap.GetTileIndex(x_max, y_max) );

	return list;
}

function Helper::ItemValuator(item)
{
	return item;
}

function Helper::ListValueSum(ai_list)
{
	local sum = 0;
	local item = ai_list.Begin();
	while(ai_list.HasNext())
	{
		sum += ai_list.GetValue(item);

		item = ai_list.Next();
	}

	return sum;
}

function Helper::CopyListSwapValuesAndItems(old_list)
{
	local new_list = AIList();
	for(local i = old_list.Begin(); old_list.HasNext(); i = old_list.Next())
	{
		local value = old_list.GetValue(i);
		new_list.AddItem(value, i);
	}

	return new_list;
}

function Helper::GetListMinValue(ai_list)
{
	ai_list.Sort(AIAbstractList.SORT_BY_VALUE, true); // highest last
	return ai_list.GetValue(ai_list.Begin());
}

function Helper::GetListMaxValue(ai_list)
{
	ai_list.Sort(AIAbstractList.SORT_BY_VALUE, false); // highest first
	return ai_list.GetValue(ai_list.Begin());
}

function Helper::Clamp(value, min, max)
{
	if(value < min)
		value = min;
	else if(value > max)
		value = max;

	return value;
}

function Helper::Min(a, b)
{
	return a < b? a : b;
}
function Helper::Max(a, b)
{
	return a > b? a : b;
}

function Helper::Abs(a)
{
	return a >= 0? a : -a;
}

function Helper::GetTileRelative(relative_to_tile, delta_x, delta_y)
{
	local tile_x = AIMap.GetTileX(relative_to_tile);
	local tile_y = AIMap.GetTileY(relative_to_tile);

	local new_x = tile_x + delta_x;
	local new_y = tile_y + delta_y;

	return AIMap.GetTileIndex(new_x, new_y);
}

function Helper::GetTileRelativeClampToMap(relative_to_tile, delta_x, delta_y)
{
	local tile_x = AIMap.GetTileX(relative_to_tile);
	local tile_y = AIMap.GetTileY(relative_to_tile);

	local new_x = Clamp(tile_x + delta_x, 0, AIMap.GetMapSizeX());
	local new_y = Clamp(tile_y + delta_y, 0, AIMap.GetMapSizeY());

	return AIMap.GetTileIndex(new_x, new_y);
}

function Helper::CanConnectToRoad(road_tile, adjacent_tile_to_connect)
{
	// If road_tile don't have road type "road" (ie it is only a tram track), then we can't connect to it
	if(!AIRoad.HasRoadType(road_tile, AIRoad.ROADTYPE_ROAD))
		return false;

	local neighbours = Helper.GetNeighbours4(road_tile);
	
	neighbours.RemoveTile(adjacent_tile_to_connect);

	// This function requires that road_tile is connected with at least one other road tile.
	neighbours.Valuate(AIRoad.IsRoadTile);
	if(Helper.ListValueSum(neighbours) < 1)
		return false;
	
	for(local neighbour_tile_id = neighbours.Begin(); neighbours.HasNext(); neighbour_tile_id = neighbours.Next())
	{
		if(AIRoad.IsRoadTile(neighbour_tile_id))
		{
			local ret = AIRoad.CanBuildConnectedRoadPartsHere(road_tile, neighbour_tile_id, adjacent_tile_to_connect);
			if(ret == 0 || ret == -1)
				return false;
		}
	}

	return true;
}

function Helper::GetNeighbours4(tile_id)
{
	local list = AITileList();

	if(!AIMap.IsValidTile(tile_id))
		return list;

	list.AddTile(Helper.GetTileRelative(tile_id, -1, 0));
	list.AddTile(Helper.GetTileRelative(tile_id, 0, -1));
	list.AddTile(Helper.GetTileRelative(tile_id, 1, 0));
	list.AddTile(Helper.GetTileRelative(tile_id, 0, 1));

	list.Valuate(AIMap.IsValidTile);
	list.KeepValue(1);

	return list;
}
